/* DupTree.cpp */
/*
()
() **************************************************************************
() * -inline   DuplicationsTree                               *
() **************************************************************************
()
*/

#include "DupTree.h"

/*  ( )  ( ) */
#define DT DuplicationsTree

/*
()       : printOn( ostream outputStream )  (  ...);
()    :         ;
()   :   ;
()  : ---.
()
*/
void DT::printOn( ostream& os ) const
  {
    Tree::printOn(os);
    if ( nOfDuplications > 0 )
      os << "_" << nOfDuplications << "_";
  }

void DT::clearNumberOfDuplications()
  {
    nOfDuplications = 0;
    for ( uint n=0; n<numberOfChildren(); n++ )
      child(n).clearNumberOfDuplications();
  }

OK_ERROR DT::defineNumberOfDuplications( RCTreeInjector injector )
  {
    clearNumberOfDuplications();  //   

    /*       */
    for ( int i=0; i<injector.injections().size(); i++ )
      if ( &RTreeInjection(injector.injections()[i]).image().root() == &root() )
        countDuplications(RTreeInjection(injector.injections()[i]));
      else
        {
          cerr << "   "
               << "DuplicationsTree::defineNumberOfDuplications()." << endl;
          return ERROR;
        }

    return OK;
  }

OK_ERROR DT::countDuplications( RTreeInjection injection )
  {
//    if ( injection.numberOfConverged() > 0 )
//      RDuplicationsTree(injection.image()).nOfDuplications ++;

    int nDup=0;
    for ( uint n=0; n<injection.numberOfChildren(); n++ )
      if ( &injection.image() == &injection.child(n).image() )
        nDup++;
    RDuplicationsTree(injection.image()).nOfDuplications += (nDup==1);

    for ( uint n=0; n<injection.numberOfChildren(); n++ )
      countDuplications( injection.child(n) );

    return OK;
  }

OK_ERROR DT::defineSumOfDuplications( RCTreeInjector injector )
  {
    clearNumberOfDuplications();  //   

    if ( defineNumberOfDuplications(injector) == ERROR )
      return ERROR;

    countSumDuplications();

    return OK;
  }

void DT::countSumDuplications()
  {
    for ( uint n=0; n<numberOfChildren(); n++ )
      {
        child(n).countSumDuplications();
        nOfDuplications += child(n).nOfDuplications;
      }
  }

OK_ERROR DT::defineNumberOfOmissions( RCTreeInjector injector )
  {
    clearNumberOfDuplications();  //   

    /*       */
    for ( int i=0; i<injector.injections().size(); i++ )
      if ( &RTreeInjection(injector.injections()[i]).image().root() == &root() )
        countOmissions(RTreeInjection(injector.injections()[i]));
      else
        {
          cerr << "   "
               << "DuplicationsTree::defineNumberOfOmissions()." << endl;
          return ERROR;
        }

    return OK;
  }

OK_ERROR DT::countOmissions( RTreeInjection injection )
  {

    if ( injection.isNotRoot() )
      {
        PDuplicationsTree p1 = PDuplicationsTree(&injection.image());
        PDuplicationsTree p0 = PDuplicationsTree(&injection.parent().image());
        if ( p1 != p0 )
          while ( ( p1 = PDuplicationsTree(&p1->parent()) ) != p0 )
            p1->nOfDuplications ++;
      }

    for ( uint n=0; n<injection.numberOfChildren(); n++ )
      countOmissions( injection.child(n) );

    return OK;
  }

OK_ERROR DT::defineSumOfOmissions( RCTreeInjector injector )
  {
    clearNumberOfDuplications();  //   

    if ( defineNumberOfOmissions(injector) == ERROR )
      return ERROR;

    countSumDuplications();

    return OK;
  }

OK_ERROR DT::defineSumOfLosses( RCTreeInjector injector )
  {
    clearNumberOfDuplications();  //   

    /*       */
    for ( int i=0; i<injector.injections().size(); i++ )
      if ( &RTreeInjection(injector.injections()[i]).image().root() == &root() )
        countDuplications(RTreeInjection(injector.injections()[i]));
      else
        {
          cerr << "   "
               << "DuplicationsTree::defineNumberOfDuplications()." << endl;
          return ERROR;
        }
    for ( int i=0; i<injector.injections().size(); i++ )
      if ( &RTreeInjection(injector.injections()[i]).image().root() == &root() )
        countOmissions(RTreeInjection(injector.injections()[i]));
      else
        {
          cerr << "   "
               << "DuplicationsTree::defineNumberOfOmissions()." << endl;
          return ERROR;
        }

    countSumLosses();

    return OK;
  }

void DT::countSumLosses()
  {
    for ( uint n=0; n<numberOfChildren(); n++ )
      {
        child(n).countSumDuplications();
        nOfDuplications += child(n).nOfDuplications;
      }
  }

RDuplicationsTree DT::operator=( RCDuplicationsTree aDupTree )
  {
    SpeciesTree::operator=(aDupTree);

    nOfDuplications = aDupTree.nOfDuplications;

    return ME;
  }

RTree DT::operator=( RCTree aTree )
  {
    SpeciesTree::operator=(aTree);

    if ( aTree.isA() == duplicationsTreeClass )
      nOfDuplications = treeToDT(aTree).nOfDuplications;

    return ME;
  }

#undef DT

