/* SpecTree.h */

#ifndef __SpecTree_h
#define __SpecTree_h

#include "BioTree.h"
#include "SrtArray.h"

_CLASSDEF(SpeciesTree)

/*   */
class SpeciesTree : public BioTree
  {

  public :

    SpeciesTree( RCString      aName="" ) : BioTree(aName)  {}
    SpeciesTree( RCSpeciesTree spTree   ) : BioTree(spTree) {}
    SpeciesTree( RCTree        aTree )    : BioTree(aTree)  {}
    ~SpeciesTree() {}

    virtual bool isSpeciesTree() const { return true; }

    bool          hasSpecies( RCString speciesName ) const;
    RSpeciesTree  species( RCString speciesName )
      { return RSpeciesTree(RCSpeciesTree(ME).species(speciesName)); }
    RCSpeciesTree species( RCString speciesName ) const;

    RSpeciesTree child( uint number )
      { return RSpeciesTree(BioTree::child(number)); }
    RCSpeciesTree child( uint number ) const
      { return RCSpeciesTree(BioTree::child(number)); }

    /*    Object'a */
    virtual classType   isA    () const { return speciesTreeClass; }
    virtual const char* nameOf () const { return "SpeciesTree"; }

    /*  "" Clonable */
    virtual Clonable* clone() const { return new SpeciesTree(ME); }

  private :

    virtual int addChild( RCTree childToAdd );
    virtual int addChild( RCString childName=""  )
      { return addChild( SpeciesTree( childName ) ); }

  };

inline int SpeciesTree::addChild( RCTree childToAdd )
  {
    if ( childToAdd.contains(ME) )
      return ERROR;

    return ( childToAdd.isA() == speciesTreeClass ?
             addChildPtr( new SpeciesTree(RCSpeciesTree(childToAdd)) ) :
             addChildPtr( new SpeciesTree(childToAdd) ) );
  }

#endif  // __SpecTree_h

