/* SrtArray.h */

#ifndef __SrtArray_h
#define __SrtArray_h

#include "ClArBase.h"  // ..  .  Clonable -
#include "Sortable.h"  // ""  

_CLASSDEF(SrtArray)

/*     Sortable */
class SrtArray : public ClArrayBase
  {

  public :

    /* //flush */
    SrtArray()                  : ClArrayBase(),
      sizeLocked(false)/*,        lastIndx(-1)*/ {}
    SrtArray( uns d )           : ClArrayBase(d),
      sizeLocked(false)/*,        lastIndx(-1)*/ {}
    SrtArray( RCSrtArray A )    : ClArrayBase(A.delta),
      sizeLocked(A.sizeLocked)/*, lastIndx(-1)*/ { ME = A; }
    SrtArray( RCClArrayBase A ) : ClArrayBase(A.delta),
      sizeLocked(false)/*,        lastIndx(-1)*/ { ME = A; }
    virtual ~SrtArray() { unlockSize(); }
    virtual void flush();

    /*    */
    bool sizeIsLocked () const { return sizeLocked; }

    /*     */
    void lockSize     ()  //  
      { sizeLocked = true;  }
    void unlockSize   ()  //     
      { sizeLocked = false; setSize(fBI); }

    /*       */
    virtual int  add( RCSortable o ) { return insertAt( o, indexFor(o) ); }
    /*        ""  */
    virtual int  removeAt( int index );

    /*    - (  i0 -   ) */
    virtual int  indexFor( RCSortable o, int i0=0 ) const;

    /*   (  i0-) -,    */
    virtual int  firstIndexOf( RCSortable o, int i0=0 ) const;

    /*        */
    /*   */ bool isEqual( RCSrtArray A ) const;

    /*   */
    RSrtArray  operator =  ( RCSrtArray A )
      { ClArrayBase::operator=(A); sizeLocked=A.sizeLocked; return ME; }
    RSrtArray  operator =  ( RCClArrayBase A );
    RSortable  operator [] ( int i )
      { return RSortable(ClArrayBase::operator[](i)); }
    RCSortable operator [] ( int i ) const
      { return RCSortable(ClArrayBase::operator[](i)); }

    /*** ,   ClArrayBase ***/

    /*    */
    virtual bool contains( RCClonable o ) const
      { return ( firstIndexOf(o) < fBI ); }

    /* min{ i: i>=i0 e(i)==o } */
    virtual int  firstIndexOf( RCClonable o, int i0=0 ) const
      { return o.isSortable() ? firstIndexOf(RCSortable(o),i0) : fBI; }

    /* / - */
    virtual RCClonable firstElement() const
      { return elementAt(0); }
    virtual RCClonable lastElement() const
      { return elementAt(fBI-1); }

    /*   */
    virtual int  add( RCClonable o )
      { return o.isSortable() ? add(RCSortable(o)) : ERROR; }

    /*    , /  */
    virtual int  firstElementIndex () const { return 0;     }
    virtual int  lastElementIndex  () const { return fBI-1; }

    /***  ***/

    /*    Object'a */
    virtual classType   isA     () const { return srtArrayClass; }
    virtual const char* nameOf  () const { return "SrtArray"; }
    virtual bool        isEqual ( RCObject o ) const
      { return &o==this || o.isA()==srtArrayClass && isEqual(RCSrtArray(o)); }
    virtual void        printOn ( ostream& os ) const;

    /*  "" Clonable */
    virtual PClonable clone() const { return new SrtArray(ME); }

  protected :

    /*  -    */
    int  insertAt ( RCSortable o, int i );

  private :

    RCSortable e( int i ) const { return RCSortable(*elPtr(i)); }

  private :

    bool sizeLocked;  //   

  };

/*

        SrtArray

      SrtArray       
,  "" Sortable,    
 ClArrayBase.

      

 *      ( ""),   0 
    lastIndex .
 *     , .. !(e(i+1)<e(i)).
 *     :
   ) ( )    (sizeLocked==false):  
        , (lastIndex+1)==size() ;
   )   (sizeLocked==ture):   size()==const,
       - (lastIndex+1)   0  size();  
      <=(size()-1)  ,  -- "".

*/

#endif  // __SrtArray_h

