/* TInject.cpp -    TreeInjector  TreeInjection */

#include "TInject.h"   //   TreeInjector  TreeInjection
#include "TreeOut.h"   //    (   log-)
#include "LogOut.h"

/*
()
() **************************************************************************
() * -inline   TreeInjector                                   *
() **************************************************************************
()
*/

TreeInjector::TreeInjector( RCClArray sources, RCSpeciesTree target ) :
  nOfSrcs(sources.size()),
  injs(nOfSrcs),
  srcs(sources),
  trgt(target),
  done(false)
  {
//tlog << "TreeInjector::TreeInjector() {" << endl;
    IMPORT_REAL( costRatio, "CostRatio" );
    for ( int i=0; i<nOfSrcs; i++ )
{
//tlog << "  N" << i << endl;
      injs.addAt( TreeInjection(RGeneTree(srcs[i]),costRatio), i );
//tlog << "ף   N" << i << endl;
}
//tlog << "} // TreeInjector::TreeInjector()" << endl;
  }

/*
()       : inject()  ();
()    :        ;
()   : ---;
()  :   / ERROR.
()
*/
double TreeInjector::inject()
  {
    double quality=0;

    for ( int i=0; i<nOfSrcs; i++ )
      {
        if ( injection(i).injectTo(trgt) == ERROR )
          {
            if ( GLOBAL_BOOLEAN("LogAllowed") )
              {
                tlog << "TreeInjector:     \n";
                TreePrinter().psGraphPrint( injection(i), tlog );
              }
            return ERROR;
          }
        quality += injection(i).quality();
      }
            
    return quality;
  }

void TreeInjector::reset()
  {
    for ( int i=0; i<nOfSrcs; i++ )
      injection(i).reset();
  }

bool TreeInjector::injectionsComplete () const
  {
    for ( int i=0; i<nOfSrcs; i++ )
      if ( injection(i).complete() == false )
        return false;
    return true;
  }

/*   ң       */
OK_ERROR TreeInjector::defineLengths()
  {
    if ( ! injectionsComplete() )
      inject();

    setNullLengths( RSpeciesTree(trgt) );
    for ( int i=0; i<nOfSrcs; i++ )
      CHECK( injection(i).defineLengths(), "" )

    return OK;
  }

void TreeInjector::setNullLengths( RSpeciesTree sTree )
  {
    sTree.distanceFromParent(0);
    for ( uint n=0; n<sTree.numberOfChildren(); n++ )
      setNullLengths(sTree.child(n));
  }

/*
()
() **************************************************************************
() * -inline   TreeInjection                                  *
() **************************************************************************
()
*/

/*
()       : inject( Tree source, SpeciesTree target )  ();
()    :  ()    ;
()   :    (  );
()  :   / ERROR.
()
*/
double TreeInjection::injectTo( RCSpeciesTree target )
  {
    if ( hasImage() )  //   ̣ 
      reset();         //   

    nOfPassed = nOfCnvrgd = sumPassed = sumCnvrgd = 0;  //   

    if ( isLeaf() )                 //  
      return injectLeaves(target);  //   

    { //  ()  
      for ( uint n=0; n<numberOfChildren(); n++ ) //    
        if ( child(n).hasImage() == false )  //   ݣ  
          CHECK( child(n).injectTo(target), "" )  //  
    } //  

    const int nOfChildren=numberOfChildren(); //   
    { //     
      TreeManipulator treeMan( child(0).image() );  //  
      for ( int n=1; n<nOfChildren; n++ )  //   ,   2-
        treeMan.goToCA( child(n).image() );  //    
      img = PSpeciesTree(&treeMan.host());  //    
    } //  

    { //      /   
      TreeManipulator treeMan(image());
      for ( int n=0; n<nOfChildren; n++ )  //    
        {
          int x=TreeManipulator( child(n).image() ).goToCA(treeMan);
          if ( x == 0 )  //   
            if ( nOfCnvrgd == 0 )  //  ݣ   
              nOfCnvrgd+=child(n).distanceFromParent();  // ++ 
            else  //   
              nOfCnvrgd = -1;  //     ( .-)
          else
            nOfPassed+=(x-1)*child(n).distanceFromParent(); // ++ 
        }
    } //  /

    if ( nOfCnvrgd == -1 )  //     ,
      nOfCnvrgd = 0;        // ,      :(

    { //     /   
      sumPassed = nOfPassed;
      sumCnvrgd = nOfCnvrgd;
      for ( int n=0; n<nOfChildren; n++ )
        sumPassed += child(n).sumOfPassed(),
        sumCnvrgd += child(n).sumOfConverged();
    } //   /  

    return quality();  //  "" 
  }

/*
()       : reset()  (?);
()    :  ;
()   : ---;
()  : ---.
()
*/
void TreeInjection::reset()
  {
    if ( isLeaf() )  //  
      return;        //   ( )
    for ( uint n=0; n<numberOfChildren(); n++ )  //   
      child(n).reset();
    img = NULL;  //  
    nOfPassed = nOfCnvrgd = sumPassed = sumCnvrgd = 0;  //   
  }

/*   ң       */
OK_ERROR TreeInjection::defineLengths()
  {
    if ( hasImage() == false )
      return ERROR;  //   ̣

    if ( isNotRoot() )
      {
        if ( ! parent().hasImage() )
          return ERROR;  //    ̣

        /*  ң         */
        uint nOfRibs = image().level() - parent().image().level();

        /*        ң,
                  ţ  */
        TreeManipulator treeMan( image() );
        for ( uint i=0; i<nOfRibs; i++, treeMan.goToParent() )
          RBioTree(treeMan.host()).distanceFromParent(
            RBioTree(treeMan.host()).distanceFromParent() +
            distanceFromParent() / nOfRibs );
      }

    /*     */
    for ( uint n=0; n<numberOfChildren(); n++ )
      CHECK( child(n).defineLengths(), "" )

    return OK;
  }


/************************************************/
//double TreeInjection::numberOfPassed() const
//  {
//    double nPassed=sumPassed;
//    for ( int i=numberOfChildren()-1; i>=0; i-- )
//      nPassed -= child(i).sumPassed;
//    return nPassed;
//  }
//double TreeInjection::numberOfConverged() const
//  {
//    double nCnvrgd=sumCnvrgd;
//    for ( int i=numberOfChildren()-1; i>=0; i-- )
//      nCnvrgd -= child(i).sumCnvrgd;
//    return nCnvrgd;
//  }
/************************************************/

/*
()       : printOn( ostream outputStream )  (  ...);
()    :         ;
()   :   ;
()  : ---.
()
*/
void TreeInjection::printOn( ostream& os ) const
  {
    GeneTree::printOn(os);
    if ( hasImage() )
      os << "{-" << numberOfPassed() << "," << numberOfConverged() << "->"
         << image().name() << "}";
  }

/******************************************************************************/

/*
 *       ( !);
 *      ,   
 *     , ..    .
 */
int TreeInjection::injectLeaves( RCSpeciesTree target )
  {
    //    (     )
    TreeManipulator tM0(ME);
    TreeManipulator tM1((RTree)target);
    //     ""   
    tM0.goToRoot();  tM0.initialize();
    tM1.goToRoot();  tM1.initialize();

    //    (    )
    const int nL0 = tM0.numberOfLeaves();
    const int nL1 = tM1.numberOfLeaves();
    if ( nL0 < 3 )  //     ң 
      {  //   -   
        if ( GLOBAL_BOOLEAN("LogAllowed") )
          {
            tlog << "TreeInjection: :  " << endl;
            TreePrinter().psGraphPrint(root(),tlog);
            tlog << " ң  -  ." << endl;
          }
        return ERROR;
      }
    if ( nL1 < 3 )  //     ң 
      {  //   -   
        if ( GLOBAL_BOOLEAN("LogAllowed") )
          {
            tlog << "TreeInjection: :  " << endl;
            TreePrinter().psGraphPrint(img->root(),tlog);
            tlog << " ң  -  ." << endl;
          }
        return ERROR;
      }

#define RETURN_ERROR_LOCAL { delete lP0; delete lP1; return ERROR; }

    //     
    PPTreeInjection lP0 = new PTreeInjection[nL0];
    PPSpeciesTree   lP1 = new PSpeciesTree[nL1];
    //      
    for ( int i=0; i<nL0; i++ )
      if ( tM0.goToNextLeaf() == ERROR )  //   < 
        RETURN_ERROR_LOCAL  //  
      else
        lP0[i] = &treeToTI(tM0.host());
    for ( int j=0; j<nL1; j++ )
      if ( tM1.goToNextLeaf() == ERROR )  //   < 
        RETURN_ERROR_LOCAL  //  
      else
        lP1[j] = PSpeciesTree(&tM1.host());
    // 

    //     
    for ( int i=0; i<nL0; i++ )
      {
        int j;
        for ( j=0; j<nL1; j++ )
          if ( lP1[j]->name()==lP0[i]->species() )
            {
              lP0[i]->img = lP1[j];
              break;
            }
        if ( j == nL1 )
          {
            if ( GLOBAL_BOOLEAN("LogAllowed") )
              tlog << endl
                   << "TreeInjection: :      \""
                   << lP0[i]->name() << "\"  " << endl << target << "\n\n";
            RETURN_ERROR_LOCAL
          }
      }  // 

#undef RETURN_ERROR_LOCAL

    delete lP0;
    delete lP1;
    return OK;
  }


RTreeInjection TreeInjection::operator=( RCTreeInjection anotherTI )
  {
    GeneTree::operator=(anotherTI);

    img       = anotherTI.img;
    nOfPassed = anotherTI.nOfPassed;
    nOfCnvrgd = anotherTI.nOfCnvrgd;
    sumPassed = anotherTI.sumPassed;
    sumCnvrgd = anotherTI.sumCnvrgd;
    if ( distanceFromParent() != anotherTI.distanceFromParent() )
      cerr << ":     ң." << endl;

    return ME;
  }

RTree TreeInjection::operator=( RCTree anotherTree )
  {
    GeneTree::operator=(anotherTree);

    if ( anotherTree.isA() == treeInjectionClass )
      {
        RCTreeInjection aT=RTreeInjection(anotherTree);
        img       = aT.img;
        nOfPassed = aT.nOfPassed;
        nOfCnvrgd = aT.nOfCnvrgd;
        sumPassed = aT.sumPassed;
        sumCnvrgd = aT.sumCnvrgd;
        if ( distanceFromParent() != aT.distanceFromParent() )
          cerr << ":     ң." << endl;
      }

    return ME;
  }

