/* TUtils.cpp */

#include "TUtils.h"

#include <math.h>
#include <time.h>

#include "LogOut.h"
#include "TreeMan.h"
#include "GlSttngs.h"

#define MYRANDOM true
// *** Random number generator ***************************************
#ifdef MYRANDOM

#define	IA   16807
#define	IM   2147483647
#define	AM   (1.0/IM)
#define	IQ   127773
#define	IR   2836
#define	MASK 123459876

/* "Minimal" random number generator of Park and Miller. Returns a uniform
   random deviate between 0.0 and 1.0 (exclusive). Set or reset idum to any
   integer value (except the unlikely value MASK) to initialize the sequence;
   idum must not be altered between calls for sussessive deviates in a sequence.
   The period of ran0 is 2^31-2 (~=2.1E9) */
float my_random01( long *idum )
{
  long k;
  float ans;
  *idum ^= MASK;                 /* XORing with MASK allows use of 0 and other */
  k = (*idum)/IQ;                /*  simple bit patterns for idum */
  *idum = IA*(*idum-k*IQ)-IR*k;  /* Compute idum=(IA*idum)%IM without overflows */
  if ( *idum<0 ) *idum += IM;    /*  by Schrage's method */
  ans = AM*(*idum);              /* Conver idum to floating result */
  *idum ^= MASK;                 /* Unmask before return */
  return ans;
}

long myrand_idum=0x55555555;

void my_srandom( long rns )
{
  myrand_idum = rns;
}

int my_random()
{
  return int(my_random01(&myrand_idum)*(RAND_MAX+1.0)-1);
}

#else
#define my_srandom srandom
#define my_random random
#endif
// *******************************************************************

static void   randomDevide( RSpeciesTree sTree );
static double meanBrancheLength( RGeneTree gTree );
static void   normalizeBrancheLengths( RGeneTree gTree,
                                       double meanLength, double normFactor );

/*    ң   */
void equalizeLengths( RGeneTree gTree )
  {
    gTree.distanceFromParent(1);
    for ( uint n=0; n<gTree.numberOfChildren(); n++ )
      equalizeLengths( gTree.child(n) );
  }

/*    ң   */
void equalizeLengths( RCClArray gTrees )
  {
    if ( GLOBAL_BOOLEAN( "LogAllowed" ) )
      tlog << " *        ." << endl;
    for ( int i=0; i<gTrees.size(); i++ )
      equalizeLengths( RGeneTree(gTrees[i]) );
    if ( GLOBAL_BOOLEAN("LogAllowed") )
      tlog << endl << " ң     ." <<endl;
  }

/*          */
void generateSpeciesTree( RCClArray gTrees, RSpeciesTree sTree )
  {

    IMPORT_BOOLEAN( logAllowed, "LogAllowed" );

    /*        */

    if ( logAllowed )
      tlog << " *    ." << endl;

    SrtArray species;
    for ( int n=0; n<gTrees.size(); n++ )
      {
        SrtArray sp=(RCGeneTree(gTrees[n])).speciesList();
        for ( int i=0; i<sp.size(); i++ )
          if ( !species.contains(sp[i]) )
            species.add(sp[i]);
      }
    if ( logAllowed )
      {
        tlog << endl << "  :" << endl;
        for ( int i=0; i<species.size(); i++ )
          tlog << (i>0?(!(i%10)?",\n":", "):"") << species[i];
        tlog << '.' << endl;
      }
    sTree=SpeciesTree();
    for ( int i=0; i<species.size(); i++ )
      TreeManipulator(sTree).addChild(RCString(species[i]));

  }

/* ""  ң   */
bool normalizeGeneTrees( RCClArray gTrees )
  {
    bool changesMade=false;
    IMPORT_BOOLEAN( logAllowed, "LogAllowed" );
    IMPORT_REAL( normFactor, "GeneTreeNormFactor" );
    for ( int i=0; i<gTrees.size(); i++ )
      {
        RGeneTree gTree=RGeneTree(gTrees[i]);
        double scaleFactor=gTree.distanceFromParent();
        if ( scaleFactor>0 && scaleFactor!=1 )
          {
            changesMade = true;
            gTree.scaleLengths( scaleFactor );
            gTree.distanceFromParent(1);
            if ( logAllowed )
              if ( scaleFactor < 1 )
                tlog << " ң " << (i+1)
                     << "-     "
                     << 1/scaleFactor << " ()." << endl;
              else
                tlog << " ң " << (i+1)
                     << "-     "
                     << scaleFactor << " ()." << endl;
          }

        if ( normFactor > 0 )
          {
            double meanLength=meanBrancheLength(gTree);
            normalizeBrancheLengths( gTree, meanLength, normFactor );
          }
      }

    return changesMade;
  }
double meanBrancheLength( RGeneTree gTree )
  {
    double meanLength=0;
    TreeManipulator treeMan(gTree);
    treeMan.goToFirstLeaf();
    do
      meanLength += RGeneTree(treeMan.host()).distanceFromParent();
    while ( treeMan.goToNextLeaf() != ERROR );
    meanLength /= gTree.numberOfLeaves();
    return meanLength;
  }
void normalizeBrancheLengths( RGeneTree gTree, double l0, double mu )
  {
    TreeManipulator treeMan(gTree);
    treeMan.goToFirstLeaf();
    do
      {
        RGeneTree gT=(RGeneTree)(treeMan.host());
        double l=gT.distanceFromParent();
        if ( l > l0 )
          gT.distanceFromParent( l0 + (l-l0)*exp(-(l/l0)*log(1+mu)) );
//          gT.distanceFromParent( l0 + (l-l0)*exp(-(l/l0)*log1p(mu)) );
//        // log1p(x) := log(1+x)
      }
    while ( treeMan.goToNextLeaf() != ERROR );
  }

/*     */
void calculateProximity( RIntMatrix proximity, RCClArray gTrees,
                         RCGenesTable gTable )
  {
    if ( GLOBAL_BOOLEAN( "LogAllowed" ) )
      tlog << " *        ."
           << endl;

    proximity.initialize(); //  - 

    /*   ͣ  */
    ClArray species(gTable.numberOfSpecies());
    for ( int a=0; a<gTable.numberOfSpecies(); a++ )
      species.addAt(gTable.species(a),a);

    /*    (  ) */
    for ( int t=0; t<gTrees.size(); t++ )
      {
        //        
        RCGeneTree gTree=RCGeneTree(gTrees[t]);
        //         
        SrtArray speciesInGTree=gTree.speciesList();
        for ( uns a=0; a<proximity.M; a++ )
          {
            //    a    
            if ( !speciesInGTree.contains( species[a] ) )
              continue;
            //    a    
            proximity(a,a) ++;
            for ( uns b=0; b<a; b++ )
              {
                //    b    
                if ( !speciesInGTree.contains( species[b] ) )
                  continue;
                TreeManipulator A(RGeneTree(gTree.species(RCString(species[a]))));
                TreeManipulator B(RGeneTree(gTree.species(RCString(species[b]))));
                proximity(a,b) += ( A.distanceFrom(B) <= 2 );
              }
          }
      }
    //     ( )
    for ( uns a=0; a<proximity.M; a++ )
      for ( uns b=a+1; b<proximity.N; b++ )
        proximity(a,b) = proximity(b,a);
  }

/*          */
/* !!!      !!! */
void orderSpeciesTreeBranches( RSpeciesTree sTree)
  {
    /*   ;) */
    if ( sTree.numberOfChildren() < 2 )
      return;

    /*     */
    orderSpeciesTreeBranches( sTree.child(0) );
    orderSpeciesTreeBranches( sTree.child(1) );

    /*  TreeMan'      */
    TreeManipulator tM0(sTree.child(0));  tM0.goToFirstLeaf();
    TreeManipulator tM1(sTree.child(1));  tM1.goToFirstLeaf();

    /*    */
    if ( tM1->name() < tM0->name() )
      tM0.changeHost(sTree.child(0)), tM0.goToParent(),
      tM1.changeHost(sTree.child(1)), tM1.goToParent(),
      tM0.interchangeChildren(tM1);
  }

/*        */
void randomizeSpeciesTree( RSpeciesTree sTree, RCIntMatrix proximity,
                           RCGenesTable gTable )
  {

    //        
    static bool firstCall=true;
    if ( firstCall )
      {
        firstCall = false;
        if ( GLOBAL_INTEGER("RandomNumberSeed") )
          my_srandom( GLOBAL_INTEGER("RandomNumberSeed") );
        else
          my_srandom( time(NULL) );
      }

    if ( GLOBAL_BOOLEAN("UseEmpiricalProbability") )
      {

        //   ͣ 
        ClArray species(gTable.numberOfSpecies());
        for ( int a=0; a<gTable.numberOfSpecies(); a++ )
          species.addAt(gTable.species(a),a);

        //    
        SrtArray speciesInUse;
        TreeManipulator treeMan(sTree);
        treeMan.goToFirstLeaf();  //      
        do  //     
          speciesInUse.add(treeMan->name());
        while ( treeMan.goToNextLeaf() != ERROR );

        //    
        treeMan.goToRoot();  treeMan.flush();

        //      (   1-  ;-) )
        while ( speciesInUse.isNotEmpty() )
          {
            //       
            uint i=my_random()%speciesInUse.size();

            //      (<=>   )
            int a=species.firstIndexOf( speciesInUse[i] );
            if ( a==ERROR )  //     ! ( ;-) )
              {
                cerr << "     !!!" << endl;
                return;
              }

            /*       {0, ..., N_a-1},  N_a -
                ,  ()  a-  */
            int r=my_random()%proximity(a,a);
            // ,     
            int b=-1;  //    
            for ( int b_=0, sum=0; uns(b_)<proximity.N; b_++ )
              if ( (sum+=(b_!=a)*proximity(a,b_)) > r ) { b=b_; break; }

            //       
            SpeciesTree littleTree;
            treeMan.changeHost(littleTree);

            if ( b>=0 )  //   b   a
              if ( speciesInUse.contains( species[b] ) )  //   ݣ  ̣
                {
                  //  ""      
                  treeMan.addChild( RCString(species[b]) );
                  treeMan.addChild( RCString(species[a]) ); //  !
                  treeMan.goToChild();
                  speciesInUse.remove(species[b]);
                }

            treeMan->name( RCString(species[a]) );  // (    )
            speciesInUse.remove(species[a]);

            TreeManipulator(sTree).addChild( littleTree );
          }

      }
    else // ! UseEmpiricalProbability
      {  // (   )
        //         
        SpeciesTree tempTree;
        TreeManipulator treeMan(sTree);
        treeMan.goToFirstLeaf();  //      
        do  //     
          TreeManipulator(tempTree).addChild(treeMan);
        while ( treeMan.goToNextLeaf() != ERROR );
        sTree=tempTree;
      }

    randomDevide(sTree);  //   "" 

  }

static void randomDevide( RSpeciesTree sTree )
  {
    uint nOfChildren = sTree.numberOfChildren();
    if ( nOfChildren < 3 )
      return;

    uint nOfLeftChildren=my_random()%(nOfChildren-1)+1;  //  "" 
    SpeciesTree leftChild, rightChild;
    for ( uint n=0; n<nOfChildren; n++ )
      if ( n < nOfLeftChildren )
        TreeManipulator(leftChild).addChild( sTree.child(n) );
      else
        TreeManipulator(rightChild).addChild( sTree.child(n) );
    TreeManipulator treeMan(sTree);
    treeMan.flush();
    treeMan.addChild( leftChild.numberOfChildren()>1 ?
                      leftChild : leftChild.child(0) );
    treeMan.addChild( rightChild.numberOfChildren()>1 ?
                      rightChild : rightChild.child(0) );

    randomDevide( sTree.child(0) );
    randomDevide( sTree.child(1) );
  }

/*     ,   ,
           */
int removeAbsentGenes( RClArray gTrees, RCSpeciesTree sTree )
  {

    SrtArray absentGenes;  //  ,    
    int numberOfAllGenes=0;
    for ( int i=0; i<gTrees.size(); )  //     
      {
        RGeneTree gTree=RGeneTree(gTrees[i]);  //    
        TreeManipulator treeMan(gTree);        //    
        bool treeRemoved=false;                //   

        numberOfAllGenes += gTree.numberOfLeaves();

        /*       */
        if ( gTree.isTrivial() )  //     
          {
            gTrees.removeAtWithShift(i);  //   
            treeRemoved = true;           //   
            continue;                     //    
          }

        /*    */
        treeMan.goToFirstLeaf();  //      
        do  //      
          if ( RCGeneTree(treeMan.host()).species() == "" ||
               !sTree.hasSpecies(RCGeneTree(treeMan.host()).species()) )
            {
              /*        */
              if ( ! absentGenes.contains( treeMan->name() ) )
                absentGenes.add( treeMan->name() );

              treeMan.removeHost();  //  -    

              /*   "" ң () */
              if ( treeMan->numberOfChildren() == 1 )
                if ( treeMan->hasParent() )
                  {
                    RBioTree( treeMan.child()).distanceFromParent(
                      RBioTree(treeMan.child()).distanceFromParent() +
                      RBioTree(treeMan.host()).distanceFromParent() );
                      //    (   )
                    treeMan->parent().addChild(treeMan.child());
                      //    
                    treeMan.removeHost();  //   
                  }
                else // treeMan->isRoot()
                  {
                    treeMan.goToFirstChild();
                    while ( treeMan->hasChild() )
                      {
                        TreeManipulator(treeMan->parent()).addChild(
                          treeMan.firstChild() );
                        treeMan.removeChild( treeMan.firstChild() );
                      }
                    treeMan.removeHost();
                  }

              /*   ң,    */
              while ( treeMan.host().isLeaf() )
                if ( treeMan.removeHost() == ERROR )
                  {
                    gTrees.removeAtWithShift(i);
                    treeRemoved = true;
                    break; // 12.04.06
                  }
            }
        while ( !treeRemoved && treeMan.goToNextLeaf() != ERROR );

// 14.10.01
        /*    ""   */
        if ( !treeRemoved )
          if ( gTree.numberOfLeaves() < 3 )  //    ң 
            {
              gTrees.removeAtWithShift(i);  //   
              continue;                     //    
            }
          else
            i++;  //     
      }

    logOutAbsentGenes( absentGenes, numberOfAllGenes );

    return ( absentGenes.size() );

  }

OK_ERROR removeGene( RGeneTree gTree, RCString gene )
  {
    TreeManipulator treeMan(gTree);  //     

    /*       */
    if ( gTree.isTrivial() )  //     
      return ERROR;

    /*    */
    treeMan.goToFirstLeaf();  //      
    do  //      
      if ( RCGeneTree(treeMan.host()).name() == gene )
        {
          treeMan.removeHost();  //  -    

          /*   "" ң () */
          if ( treeMan->numberOfChildren() == 1 )
            if ( treeMan->hasParent() )
              {
                RBioTree( treeMan.child()).distanceFromParent(
                  RBioTree(treeMan.child()).distanceFromParent() +
                  RBioTree(treeMan.host()).distanceFromParent() );
                  //    (   )
                treeMan->parent().addChild(treeMan.child());
                  //    
                treeMan.removeHost();  //   
              }
            else // treeMan->isRoot()
              {
                treeMan.goToFirstChild();
                while ( treeMan->hasChild() )
                  {
                    TreeManipulator(treeMan->parent()).addChild(
                      treeMan.firstChild() );
                    treeMan.removeChild( treeMan.firstChild() );
                  }
                treeMan.removeHost();
              }

// 12.04.06
//          /*   ң,    */
//          while ( treeMan.host().isLeaf() )
//            if ( treeMan.removeHost() == ERROR )
//              break;

          return OK;
        }
    while ( treeMan.goToNextLeaf() != ERROR );

    return ERROR;
  }

void setSpeciesOnGeneTrees( RCClArray gTrees, RCGenesTable gTable )
  {
    for ( int i=0; i<gTrees.size(); i++ )
      {
        TreeManipulator treeMan( (RTree)(gTrees[i]) );
        treeMan.goToFirstLeaf();
        do
          RGeneTree(treeMan.host()).setSpecies(
            gTable.whatSpecies(treeMan->name()) );
        while ( treeMan.goToNextLeaf() != ERROR );
      }
  }
