/* Tree.h */

#ifndef __TREE_H
#define __TREE_H

#include "ClString.h"
#include "TreeBase.h"

_CLASSDEF(Tree)

extern const Tree nullTree;

/*  (     ) */
class Tree : public TreeBase
  {

  public :

    /* , , destroy() */
    Tree( RCString aName="", uint aDelta=2 ) :
      TreeBase( aDelta    ), _name( aName    ), _level( 0 ) {}
    Tree( RCTree   aT ) :
      TreeBase( aT._delta ), _name( aT._name ), _level( 0 ) { ME = aT; }
    virtual ~Tree() {}

    /*    Tree */

    virtual int addChild    ( RCTree   childToAdd    );  //  
    virtual int addChild    ( RCString childName=""  )   //  new 
      { return addChild( Tree( childName ) ); }
    /*   */ int removeChild ( RCString childName     )   //  
      { return TreeBase::removeChild( childNumber( childName ) ); }
    /*   */ int removeChild ( RTree    childToRemove )   //  
      { return TreeBase::removeChild( childNumber( childToRemove ) ); }

    bool isEqual( RCTree aTree ) const;

    /*    -      */
    String name( RCString newName )  { return (_name=newName);    }
    String name             () const { return  _name;             }
    uint   updateLevel();  //       
    uint   level            () const { return  _level;            }
    uint   numberOfChildren () const { return  TreeBase::numberOfChildren(); }
    uint   numberOfLeaves   () const;  //  
    uint   numberOfSubtrees () const;  //  
    uint   height           () const;  //  ()
    uint   maximumPower     () const;  //  (.)  
    uint   sizeInMemory     () const;  //    ()
    bool   hasParent        () const { return TreeBase::hasParent(); }
    bool   hasChild         () const { return TreeBase::hasChild();  }
    bool   isRoot           () const { return  _level==0;         }
    bool   isLeaf           () const { return !TreeBase::numberOfChildren(); }
    bool   isNotRoot        () const { return !isRoot();          }
    bool   isNotLeaf        () const { return !isLeaf();          }
    bool   isTrivial        () const { return isRoot()&&isLeaf(); }
    bool   contains( RCTree anotherTree ) const;  //   

    /* (      
           ,    Tree.) */
    void   becomeNULL ()       { flush(); name("<<NULL>>"); }
    bool   isNULL     () const { return ( isTrivial() && name()=="<<NULL>>" ); }

    /*    / */
    RCTree parent     () const { return RTree(TreeBase::parent()); }  // 
    RTree  parent     ()       { return RTree(TreeBase::parent()); }  // 
    RCTree firstChild () const { return child(0); }  //  
    RTree  firstChild ()       { return child(0); }  //  
    RCTree root       () const;  //    ()
    RTree  root       ()       { return RTree(RCTree(ME).root()); };

    /*       */
    int    childNumber ( RCTree   childToFind ) const { return TreeBase::childNumber(RCTreeBase(childToFind));} //   
    int    childNumber ( RCString name ) const; //   
    RCTree child       ( uint     n    ) const { return RTree(TreeBase::child(n)); }
    RTree  child       ( uint     n    )       { return RTree(TreeBase::child(n)); }
    int    removeChild ( int      n    )       { return TreeBase::removeChild(n);  }
    /* ( 5 , ,    .) */

    /*   */
    virtual RTree operator =  ( RCTree anotherTree );       // 
    /*   */ bool  operator == ( RCTree anotherTree ) const  // 
      { cerr<<"!"; return ( this == &anotherTree ); } //   ()
    /*   */ bool  operator != ( RCTree anotherTree ) const  // 
      { cerr<<"!"; return !operator==( anotherTree ); }

    /*    Object'a */
    virtual classType   isA     () const { return treeClass; }
    virtual const char* nameOf  () const { return "Tree"; }
    virtual bool        isEqual ( RCObject object ) const;
    virtual void        printOn ( ostream& os     ) const;

    /*  "" Clonable */
    virtual PClonable clone() const { return new Tree(ME); }

    /*  "" Sortable */
    virtual bool isLessThan( RCSortable s ) const
      { return ( s.isA()==isA() && false ); }

  protected :

    int addChildPtr( PTreeBase newChildPtr )
      {
        const int newChildNumber=TreeBase::addChildPtr( newChildPtr );
        child(newChildNumber).updateLevel();
        return newChildNumber;
      }

  private :

    /* -  */
    String _name;   // 
    uint   _level;  // ""     ()

    friend class TreeManipulator;  // -   Tree

  };

inline bool Tree::isEqual( RCObject object ) const
  { return ( object.isA()==isA() && isEqual(RTree(object)) ); }

#endif  // __TREE_H

