/* TreeMan.h */

#ifndef __TREEMAN_H
#define __TREEMAN_H

#include "Tree.h"
#include "ClArray.h"

/*  #define    */
#define TM TreeManipulator

/*   TreeManipulator ======================================= */

_CLASSDEF(TM)

/*   -    */
class TreeManipulator
  {

  public :

    /* ,  */
    TM( RTree tree    ) : hostPtr(&tree  ), curKidN(UNDEFINED) { initialize(); }
    TM( PTree treePtr ) : hostPtr(treePtr), curKidN(UNDEFINED) { initialize(); }
    TM( TM const & tM ) : hostPtr(tM.hostPtr), curKidN(tM.curKidN) {} // added 'const' 2006-03-03
    ~TreeManipulator() {}

    /*    TreeManipulator */

    RTree host() const { return *hostPtr; }

    void  changeHost( RTree tree ) { hostPtr = &tree, initialize(); }

    int   initialize ()
      { return ( curKidN = ( hostPtr->hasChild() ? 0 : UNDEFINED ) ); }
    void  flush      () { hostPtr->flush(); curKidN = UNDEFINED; }

    /* */ operator RTree () const { return *hostPtr; }
    PTree operator ->    () const { return  hostPtr; }

    //  /   
    int   addChild    ( RCString name="" )
      { return ( curKidN = hostPtr->addChild(name) ); }
    int   addChild    ( RCTree   aChild   )
      { return ( curKidN = hostPtr->addChild(aChild) ); }
    int   addChildren ( const RClArray children );
    int   removeChild ( RCString name    );
    int   removeChild ( RTree    child   );
    int   removeChild ();
    int   removeHost  ()
      { return ( host().hasParent() ? (goToParent(),removeChild()) : ERROR ); }

    //  
    int   findChild( RTree   childToFind )
      { return ( curKidN = hostPtr->childNumber( childToFind ) ); }
    int   findChild( RCString name        )
      { return ( curKidN = hostPtr->childNumber( name ) ); }

    //  
    int   goToParent     ();  //     
    int   goToChild      ();  //      
    int   goToFirstChild ();  //      
    int   goToNextChild  ();  //      
    void  goToRoot       ();  //     
    int   goToFirstLeaf  ();  //      
    int   goToNextLeaf   ();  //     (.. )
    int   goUpNLevels ( int n );  //   n  
    int   goToCA ( TM partner ); //    
    int   goToSubtree ( RCString name );

    RTree child      () const { return host().child( curKidN ); }
    RTree firstChild ()       { initialize(); return child(); }
    RTree nextChild  ()       { return host().child( incrCurKidN() ); }

    bool  hasMoreChild   () const;
    int   numberOfLeaves () const { return hostPtr->numberOfLeaves(); }

    RTree commonAncestor ( RCTreeManipulator partner ) const;
    int   distanceFrom   ( RCTreeManipulator partner ) const;

    void     scaleLengths         ( double scaleFactor );
    OK_ERROR interchangeChildren  ( RCTreeManipulator partner );
    OK_ERROR takeChildFrom        ( RCTreeManipulator partner );
    OK_ERROR nameUntitledSubtrees ( RCString beginWith="",
                                    RCString endWith="" );

  private :

    PTree hostPtr;
    int   curKidN;

    int zeroCurKidN() { return ( curKidN = 0 ); }
    int incrCurKidN() { return ( hasMoreChild() ? ++curKidN : zeroCurKidN() ); }

  };


/* inline-  TreeManipulator ==================================== */

/*    */
inline int TM::removeChild( RCString name  )
  { int t=host().removeChild(name);     initialize();  return t; }
inline int TM::removeChild( RTree    aChild )
  { int t=host().removeChild(aChild);   initialize();  return t; }
inline int TM::removeChild()
  { int t=host().removeChild(curKidN);  initialize();  return t; }

/*    */
inline int TM::goToParent()
  {
    if ( ! hostPtr->hasParent() )  return ERROR;
    curKidN = hostPtr->parent().childNumber(*hostPtr);
    hostPtr = &hostPtr->parent();  return OK;
  }

/*     */
inline int TM::goToChild()
  {
    if ( curKidN == UNDEFINED )
      return ERROR;
    changeHost( child() );
    return OK;
  }

/*     */
inline int TM::goToFirstChild()
  { initialize(); return goToChild(); }

/*     */
inline int TM::goToNextChild()
  { incrCurKidN(); return goToChild(); }

inline bool TM::hasMoreChild() const
  {
    return ( hostPtr->hasChild() &&
             unsigned(curKidN+1) < hostPtr->numberOfChildren() );
  }


/*     TreeManipulator'  */
inline RTree TM::commonAncestor( RCTreeManipulator partner ) const
  {
    TM treeMan(ME);
    treeMan.goToCA( RTreeManipulator(partner) );
    return treeMan.host();
  }
/*     */
inline int TM::distanceFrom( RCTreeManipulator partner ) const
  {
    TM treeMan(ME);
    return treeMan.goToCA( RTreeManipulator(partner) );
  }

#undef TM

#endif  // __TREEMAN_H

